#include <stdlib.h>
#include <stdio.h>
#include <smaky.h>
#include "png.h"
#include "memory.h"
#include "png_iface.h"

extern const char* png_get_libpng_ver();

#ifdef TEST
extern void pnglibClose();
extern void OpenPNGLib();
extern void ClosePNGLib();
#endif

/* Options de chargement. */
#define	MODE_SUPPORT1BIT	(1<<15)
#define	MODE_SUPPORT2BIT	(1<<14)
#define	MODE_SUPPORT4BIT	(1<<13)
#define	MODE_SUPPORT8BIT	(1<<12)
#define	MODE_SUPPORTRGB		(1<<11)
#define	MODE_SUPPORTALPHA	(1<<10)

/* Options communes. */
#define	MODE_BGR			(1<<6)
#define	MODE_RGBX			(1<<5)
#define	MODE_XRGB			(1<<4)

/* Options de sauvegarde. */
#define	MODE_INTERLACE		(1<<8)
#define	MODE_ALPHA			(1<<7)

#define	ERR_FILEOPEN	-1
#define	ERR_PNGCREATE	-2
#define	ERR_FAILED		-3
#define	ERR_FORMAT		-4

#define	SCREEN_GAMMA	2.20

static void PNGerror( png_struct* pPNG, char* message )
{
	L_aftext( "PNG error : ");
	L_aftext( message );
	L_aftext( "\n" );
	longjmp( pPNG->jmpbuf, 1 );
}

static void PNGwarning( png_struct* pPNG, char* message )
{
	if (!pPNG)
		return;
	L_aftext( "PNG warning : ");
	L_aftext( message );
	L_aftext( "\n" );
}

/*****************/
/* PNGdecompress */
/*****************/
/* clut ^^ buffer. if null, memory is requested */
short PNGdecompress( u_long MemType, u_long MemAccount, u_long mode, const char* name, Point *size, u_short *bpp, u_short *bpl, u_char* *buffer, u_char* *clut, u_short *nbcols )
{
	FILE* fr;
	png_struct* pPNG;
	png_info* pInfo;
	short error = 0;
	int nbpass, pass;
	short y;
	
	*clut = NULL;
	*buffer = NULL;

#ifdef MONDEBUG
L_afx8((long)name);
#endif
	
	AddProcId( MemType, MemAccount );
	
	fr = fopen( name, "rb" );
	if (!fr) {
		RemoveProcId();
		return ERR_FILEOPEN;
	}
	pPNG = png_create_read_struct( png_get_libpng_ver(), NULL, PNGwarning, PNGerror );
	if (!pPNG) {
		fclose( fr );
		RemoveProcId();
		return ERR_PNGCREATE;
	}
	pInfo = png_create_info_struct( pPNG );
	if (!pInfo) {
		png_destroy_read_struct( &pPNG, NULL, NULL );
		fclose( fr );
		RemoveProcId();
		return ERR_PNGCREATE;
	}
	if (setjmp( pPNG->jmpbuf )) {
		png_destroy_read_struct( &pPNG, &pInfo, NULL );
#if 1
		if (clut && *clut)
			free( *clut );
		if (buffer && *buffer)
			free( *buffer );
#endif
		fclose( fr );
		RemoveProcId();
		return ERR_FAILED;
	}
	png_init_io( pPNG, fr );
	png_read_info( pPNG, pInfo );
	
	size->x = pInfo->width;
	size->y = pInfo->height;
	*bpp = pInfo->bit_depth * pInfo->channels;

	if (pInfo->bit_depth==16)
		png_set_strip_16( pPNG );

	if (pInfo->color_type & PNG_COLOR_MASK_ALPHA) {
		if (pInfo->valid & PNG_INFO_bKGD)
			png_set_background( pPNG, &pInfo->background, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
		else {
			png_color_16 backgr;
			backgr.red = backgr.green = backgr.blue = backgr.gray = 0;
			png_set_background( pPNG, &backgr, PNG_BACKGROUND_GAMMA_SCREEN, 0, SCREEN_GAMMA );
		}
	}

#ifdef TEST
	printf("color_type = %x\n", pInfo->color_type );
	printf("bit_depth = %i\n", pInfo->bit_depth );
#endif

	switch (pInfo->bit_depth) {
	case 1:
		if (!(mode & MODE_SUPPORT1BIT))
			if (mode & MODE_SUPPORT8BIT)
				png_set_packing( pPNG );
			else if (mode & MODE_SUPPORTRGB)
				png_set_expand( pPNG );
			else
				error = ERR_FORMAT;
		if (pInfo->color_type==PNG_COLOR_TYPE_GRAY || pInfo->color_type==PNG_COLOR_TYPE_GRAY_ALPHA)
			png_set_invert_mono( pPNG );
		break;
	case 2:
		if (!(mode & MODE_SUPPORT2BIT))
			if (mode & MODE_SUPPORT8BIT)
				png_set_packing( pPNG );
			else if (mode & MODE_SUPPORTRGB)
				png_set_expand( pPNG );
			else
				error = ERR_FORMAT;
		break;
	case 4:
		if (!(mode & MODE_SUPPORT4BIT))
			if (mode & MODE_SUPPORT8BIT)
				png_set_packing( pPNG );
			else if (mode & MODE_SUPPORTRGB)
				png_set_expand( pPNG );
			else
				error = ERR_FORMAT;
		break;
	case 16:	/* Converted, but not updated */
		png_set_strip_16( pPNG );
	case 8:
		if (pInfo->color_type==PNG_COLOR_TYPE_RGB_ALPHA || pInfo->color_type==PNG_COLOR_TYPE_RGB) {		/* RGB */
			if (!(mode & MODE_SUPPORTRGB))
				error = ERR_FORMAT;
		} else if (!(mode & MODE_SUPPORT8BIT)) {
			if (mode & MODE_SUPPORTRGB)
				png_set_expand( pPNG );
			else
				error = ERR_FORMAT;
		}
		break;
	}
	
	if (mode & MODE_RGBX)
		png_set_filler( pPNG, 0, PNG_FILLER_AFTER );
	else if (mode & MODE_XRGB)
		png_set_filler( pPNG, 0, PNG_FILLER_BEFORE );

	if (error) {
		png_destroy_read_struct( &pPNG, &pInfo, NULL );
		fclose( fr );
		RemoveProcId();
		return error;
	}

	nbpass = png_set_interlace_handling( pPNG );
	
	png_read_update_info( pPNG, pInfo );
	
	if (nbcols)
		*nbcols = (pInfo->color_type & PNG_COLOR_MASK_PALETTE) ? pInfo->num_palette : 0;
	if ((pInfo->color_type & PNG_COLOR_MASK_PALETTE) && pInfo->num_palette && clut) {
		png_color* rd;
		u_char* wr;
		int col;
		if (!*clut)
			*clut = malloc( pInfo->num_palette * 3 );
		wr = *clut;
		if (!*clut)
			longjmp( pPNG->jmpbuf, 1 );
		rd = pInfo->palette;
		for( col=pInfo->num_palette; col; col--, rd++ ) {
			*wr++ = rd->red;
			*wr++ = rd->green;
			*wr++ = rd->blue;
		}
	}
	
	size->x = pInfo->width;
	size->y = pInfo->height;
	*bpp = pInfo->bit_depth * pInfo->channels;
	*bpl = (pInfo->width * *bpp * 2 + 15) / 16;
	*buffer = malloc( *bpl * pInfo->height );
	if (!*buffer)
		longjmp( pPNG->jmpbuf, 1 );
	
	png_start_read_image( pPNG );

	for( pass=nbpass; pass; pass-- ) {
		u_char* wr;
		wr = *buffer;
		for( y=pInfo->height; y; y-- ) {
			png_read_row( pPNG, wr, NULL );
			wr += *bpl;
		}
	}
	
	png_read_end( pPNG, pInfo );

	png_destroy_read_struct( &pPNG, &pInfo, NULL );

	fclose( fr );

	RemoveProcId();
	return 0;
}

/***************/
/* PNGcompress */
/***************/
short PNGcompress( u_long mode, const char* name, Point size, u_long bpp, u_long bpl, u_char* buffer, u_char* clut, u_short nbcols )
{
	png_struct* pPNG;
	png_info* pInfo;
	FILE* fw;
	png_color palette[256];
	int interlacePass;

#ifdef MONDEBUG
L_aftext("\r\rbpp=");
L_afx8((long)bpp);
L_aftext("\rbpl=");
L_afx8((long)bpl);
L_aftext("\r");
#endif

	fw = fopen( name, "wb" );
	if (!fw)
		return ERR_FILEOPEN;

	pPNG = png_create_write_struct( png_get_libpng_ver(), NULL, PNGwarning, PNGerror );
	if (!pPNG) {
		fclose( fw );
		return ERR_PNGCREATE;
	}
	pInfo = png_create_info_struct( pPNG );
	if (!pInfo) {
		png_destroy_write_struct( &pPNG, NULL );
		fclose( fw );
		return ERR_PNGCREATE;
	}
	
	if (setjmp( pPNG->jmpbuf )) {
		png_destroy_write_struct( &pPNG, &pInfo );
		fclose( fw );
		return ERR_FAILED;
	}
	
	png_init_io( pPNG, fw );
	
	png_set_compression_level( pPNG, 9 );
	pInfo->width = size.x;
	pInfo->height = size.y;
	pInfo->interlace_type = (mode & MODE_INTERLACE) ? 1 : 0;
	
	if (bpp==24 || bpp==32) {
		pInfo->color_type = PNG_COLOR_MASK_COLOR;
		pInfo->bit_depth = 8;
		pInfo->channels = 3;
		if (mode & MODE_RGBX)
			png_set_filler( pPNG, 0, PNG_FILLER_AFTER );
		else if (mode & MODE_XRGB)
			png_set_filler( pPNG, 0, PNG_FILLER_BEFORE );
	} else {
		pInfo->bit_depth = bpp;
		if (clut) {
			u_char* rdclut = clut;
			png_color* wrcolor = palette;
			int nb;
			if (nbcols>256)
				nbcols = 256;
			pInfo->color_type = PNG_COLOR_MASK_PALETTE | PNG_COLOR_MASK_COLOR;
			for( nb=0; nb<nbcols; nb++, rdclut+=3, wrcolor++ ) {
				wrcolor->red = rdclut[0];
				wrcolor->green = rdclut[1];
				wrcolor->blue = rdclut[2];
			}
			pInfo->num_palette = nbcols;
			pInfo->palette = palette;
			pInfo->valid |= PNG_INFO_PLTE;
		}
	}
	if (mode & MODE_ALPHA)
		pInfo->color_type |= PNG_COLOR_MASK_ALPHA;

	png_write_info( pPNG, pInfo );
	
	if (mode & MODE_BGR)
		png_set_bgr( pPNG );

	interlacePass = png_set_interlace_handling( pPNG );

	for( ; interlacePass; interlacePass-- ) {
		u_char* rdLine = buffer;
		u_short y = size.y;
		for( ; y; y-- ) {
			png_write_row( pPNG, rdLine );
			rdLine += bpl;
		}
	}

	png_write_end( pPNG, pInfo );

	png_destroy_write_struct( &pPNG, &pInfo );
	fclose( fw );
	return 0;
}

/* short PNGdecompress( u_short MemType, u_long MemAccount, u_long mode, const char* name, Point *size, u_short *bpp, u_short *bpl, u_char* *buffer, u_char* *clut ) */
/* in:	d1.16	MemType
 *		a1.32	MemAccount
 *		d3.32	mode
 *		a3.32	^name
 * out:	d2.32	size
 *		d4.16	nb cols in CLUT
 *		d5.16	bpl
 *		d6.16	bpp
 *		a4.32	buffer
 *		a5.32	clut
 *		d7.16	erreur
 */
void AsmPNGdecompress()
{
	asm volatile(	"\n"
					"	moveml	d0/d1/a0/a1,sp@-\n"
					"	subl	#18,sp\n"
					"	movel	sp,a0\n"
					"	movel	a0,sp@-\n"	/* ^nbcols */
					"	addqw	#2,a0\n"
					"	movel	a0,sp@-\n"	/* ^clut */
					"	addqw	#4,a0\n"
					"	movel	a0,sp@-\n"	/* ^buffer */
					"	addqw	#4,a0\n"
					"	movel	a0,sp@-\n"	/* ^bpl */
					"	addqw	#2,a0\n"
					"	movel	a0,sp@-\n"	/* ^bpp */
					"	addqw	#2,a0\n"
					"	movel	a0,sp@-\n"	/* ^size */
					"	movel	a3,sp@-\n"	/* name */
					"	movel	d3,sp@-\n"	/* mode */
					"	movel	a1,sp@-\n"	/* MemAccount */
					"	extl	d1\n"
					"	movel	d1,sp@-\n"	/* MemType */
					"	jbsr	_PNGdecompress\n"
					"	addl	#40,sp\n"
					"	movew	sp@+,d4\n"
					"	movel	sp@+,a5\n"
					"	movel	sp@+,a4\n"
					"	movew	sp@+,d5\n"
					"	movew	sp@+,d6\n"
					"	movel	sp@+,d2\n"
					"	movew	d0,d7\n"
					"	moveml	sp@+,d0/d1/a0/a1\n"	);
}

/*short PNGcompress( u_long mode, const char* name, Point size, u_short bpp, u_short bpl, u_char* buffer, u_char* clut )*/
/* in:	d2.32	size
 *		d3.32	mode
 *		d4.16	nb cols in CLUT
 *		d5.16	bpl
 *		d6.16	bpp
 *		a3.32	^name
 *		a4.32	buffer
 *		a5.32	clut
 * out:	d7.16	erreur
 */
void AsmPNGcompress()
{
	asm volatile(	"\n"
					"	moveml	d0/d1/a0/a1,sp@-\n"
					"	movew	d4,sp@-\n"
					"	movel	a5,sp@-\n"
					"	movel	a4,sp@-\n"
					"	extl	d5\n"
					"	movel	d5,sp@-\n"
					"	extl	d6\n"
					"	movel	d6,sp@-\n"
					"	movel	d2,sp@-\n"
					"	movel	a3,sp@-\n"
					"	movel	d3,sp@-\n"
					"	jbsr	_PNGcompress\n"
					"	addl	#30,sp\n"
					"	movew	d0,d7\n"
					"	moveml	sp@+,d0/d1/a0/a1\n"	);
}


#ifdef TEST
#include <gra2.h>
int main( int argc, char** argv )
{
	Point size = {0,0};
	u_short bpp=0, bpl=0;
	u_char* buffer = NULL;
	u_char* clut = NULL;
	int err;
	u_long mode;
	u_short nbcols;
	
	Windesc display;
	gra2_open( &display );
	OpenPNGLib();
	
	mode = MODE_SUPPORT1BIT;
	switch (display->dfcnp) {
	case 4:
		mode|= MODE_SUPPORT4BIT;
		break;
	case 8:
		mode = MODE_SUPPORT8BIT;
		break;
	}
	err = PNGdecompress( 0, 0, mode, (argc==2) ? argv[1] : "#MM0:C:CLIBS:PNG:PNGLIV.PNG", &size, &bpp, &bpl, &buffer, &clut, &nbcols );
	printf("PNGdecompress = %i\n", err );
	printf("size=(%i,%i) bpp=%i bpl=%i buffer=%x clut=%x nbcols=%i\n", size.x, size.y, bpp, bpl, buffer, clut, nbcols );
	if (!err || (buffer && bpp && bpl)) {
		err = PNGcompress( 0, "test.png", size, bpp, bpl, buffer, clut, nbcols );
		printf("PNGcompress = %i\n", err );
		printf("Hit a key to display image\n");
		L_getcar();
		gra2_craster( display, (Point){0,0}, buffer, bpl, (Point){0,0}, NULL, 0, size, LOADDOT|((display->dfcmd)?DSTCOUL:0)|((bpp>1)?SRCCOUL:0) );
	}
	
	L_getcar();
	gra2_close();
	ClosePNGLib();
	pnglibClose();
	return 0;
}
#endif
